/*
  Firmware for Bitmap Projector
  Copyright (C) 2011 Matt Evans

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


  JVC/Daewoo IR remote driver code:

  Input is inverted.  Base state is 1, then we see a 9ms low time,
  then 4ms high time.  Bits follow; an edge 1->0 is the start; stays 0
  for 600us, then high for 550us for a 0 bit, and
  0 for 600us then high for 1600us for 1 bit.
  Transmissions are 60ms apart, min.
  7 addr bits, 7 cmd bits.

  Do this using thresholds; in idle state, spot the input 9.5ms pulse
  by waiting for 1->0 then at 0->1 checking time; if between
  ST_A_MIN/ST_A_MAX move to next state waiting for 1->0 again; if
  this happens between ST_B_MIN/ST_B_MAX move to wait for bits.
  (Actually, fold this into below, the 4ms 1 can look like a 1 bit of
  data. ) When waiting for bits, on 1->0 transition check for timeout
  (1 time was > 5ms).  if it was <5ms but >~1000, it's a 1. If it was
  < ~1000, it's a 0.  on next 0->1 check 0 time was 600us.
   
  Each time shift bits into a uint32.  Info on JVC proto says 8 bits 
  addr, 8 bits data-- but I see changes only in a /second/ 16 bits.
*/

#include "hw.h"
#include <inttypes.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <stdio.h>
#include "serio.h"
#include "ir.h"

static uint8_t t1_ovfs;

#define IR_IDLE 0
#define IR_IN_STARTBIT 1
#define IR_DATA 2
static uint8_t ir_bits;
static uint8_t ir_state;
static uint8_t ir_err;

volatile uint32_t ir_rx;
volatile uint8_t ir_new_code;

#define US(x) (16*(x))
#define IR_START_BIT_MIN        US(7000UL)
#define IR_START_BIT_MAX        US(11000UL)
#define IR_DBIT_0_MIN           US(450UL)
#define IR_DBIT_0_MAX           US(700UL)
#define IR_DBIT_1_MIN           US(450UL)
#define IR_DBIT_1_ZEROONE_THRESH US(1000UL)
#define IR_DBIT_1_MAX           US(5000UL)

static void ir_abort(void)
{
  PS("abt\r\n");
  /* Reset the FSM stuff, bad data detected. */
  ir_state = IR_IDLE;
  LED_OFF();
}

/* When IR pin changes, this is called. */
ISR(PCINT0_vect)
{
  uint32_t time_now;

  time_now = ((uint32_t)t1_ovfs) * 65536UL + (uint32_t)TCNT1;

  if (PINB & B(PB2)) {
    /* 0->1 transition */
    switch (ir_state) {
      case IR_IDLE:
        /* Spurious! */
        break;
      case IR_IN_STARTBIT:
        /* End of start bit.  Was it the right length? */
        if ((time_now > IR_START_BIT_MAX) ||
            (time_now < IR_START_BIT_MIN)) {
          /* Spurious bad start bit; ignore. */
          PS("Bad start bit, len ");
          phex32(time_now);
          putchr(' ');
          phex32(IR_START_BIT_MIN);
          putchr(' ');
          phex32(IR_START_BIT_MAX);
          pnewl();
          ir_abort();
          ir_err |= B(0);
        } else {
          /* Good size! */
          ir_state = IR_DATA;
          ir_bits = 0;
          ir_rx = 0;
        }
        break;
      case IR_DATA:
        /* How long was it 0 for? */
        if ((time_now > IR_DBIT_0_MAX) ||
            (time_now < IR_DBIT_0_MIN)) {
          /* Rubbish */
          PS("Bad data bit, len ");
          phex32(time_now);
          putchr(' ');
          phex32(IR_DBIT_0_MIN);
          putchr(' ');
          phex32(IR_DBIT_0_MAX);
          pnewl();
          ir_abort();
          ir_err |= B(1);
        } else {
          /* OK, but do nothing; the 1-time determines the data */
        }
        break;
      default:
        PS("Bad 01 state ");
        phex8(ir_state);
        PS(", time ");
        phex32(time_now);
        pnewl();
        /* Impossible lunacy */
        ir_abort();
        ir_err |= B(3);
        break;
    }
  } else {
    /* 1->0 transition */
    switch (ir_state) {
      case IR_IDLE:
        /* Start bit now? */
        if (!ir_new_code) {
          LED_ON();
          ir_state = IR_IN_STARTBIT;
        }
        break;
      case IR_DATA:
        if ((time_now > IR_DBIT_1_MAX) ||
            (time_now < IR_DBIT_1_MIN)) {
          /* Rubbish */
          ir_abort();
          ir_err |= B(2);
        } else {
          /* Yay! A data bit! */
          if (time_now > IR_DBIT_1_ZEROONE_THRESH) {
            /* Bit 1 */
            ir_rx = (ir_rx >> 1) | 0x80000000UL;
          } else {
            /* Bit 0 */
            ir_rx = (ir_rx >> 1);
          }
          if (++ir_bits >= 32) {
            /* Complete! */
            ir_state = IR_IDLE;
            ir_new_code = 1;
            LED_OFF();
          }
          /* Else stay in IR_DATA state */
        }
        break;
      default:
        /* Can't be IR_IN_STARTBIT */
        PS("Bad 10 state ");
        phex8(ir_state);
        PS(", time ");
        phex32(time_now);
        pnewl();
        ir_abort();
        ir_err |= B(3);
        break;
    }
  }
  /* Reset timer to time until next change */
  t1_ovfs = 0;
  TCNT1 = 0;
}

ISR(TIMER1_OVF_vect)
{
  if (t1_ovfs < 255)
    t1_ovfs++;
  /* Sanity check: Timer is reset whenver there's a change observed on
     the IR input.  If 30ms have passed since the last change, reset.
     This is wayyy over the bit time. */
#if 1
  if ((ir_state != IR_IDLE) && (t1_ovfs > (30000UL*16/65536))) {
    /* Timeout */
    t1_ovfs = 0;
    ir_abort();
  }
#endif
}

void ir_init(void)
{
  PORTB |= B(PB2); /* Pullup */
  DDRB &= B(PB2);  /* IR input (Also PCINT2) */

  PCMSK0 |= B(PCINT2);
  PCIFR = 0x07;
  PCICR |= B(PCIE0);

  t1_ovfs = 0;
  /* Plain counter mode */
  TCCR1A = 0;
  TCCR1C = 0;
  TCNT1 = 0;
  TIMSK1 = B(TOIE1); /* Ovf at 0xffff */
  TCCR1B = B(CS10); /* Go! at 20MHz, 20 clocks per us */

  ir_state = IR_IDLE;
  ir_bits = 0;
  ir_rx = 0;
  ir_err = 0;
  ir_new_code = 0; /* This flags newness */
}

uint8_t ir_code_to_button(uint32_t c)
{
  switch (c >> 16) {
    case 0xda25:
      return BTN_PWR;
      break;
    case 0xc837:
      return BTN_MUTE;
      break;
    case 0xd42b:
      return BTN_AUTO;
      break;
    case 0xd22d:
      return BTN_DISP;
      break;
    case 0xe619:
      return BTN_VOLMINUS;
      break;
    case 0xf40b:
      return BTN_VOLPLUS;
      break;
    case 0xfc03:
      return BTN_ESC;
      break;
    case 0xfa05:
      return BTN_MENU;
      break;
    case 0x629d:
      return BTN_PCAV;
      break;
  }
  return BTN_UNK;
}
