/*
  Firmware for Bitmap Projector
  Copyright (C) 2011 Matt Evans

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

  Driver code for Nokia 1200 96x68 LCD.
  The particular LCD that I used seems to diverge from what datasheets
  suggest; I've left in commented-out commands that don't seem to be necessary
  (or don't work) on this LCD but may be useful for others.
*/

#include "hw.h"
#include <inttypes.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <stdio.h>
#include "serio.h"
#include "lcd.h"

#define LCD_NRESET      B(PC2)
#define LCD_DIN         B(PC3)
#define LCD_CLK         B(PC0)
#define LCD_NCS         B(PC1)

/* Max CLK freq is 6.5MHz.  At 16MHz CPU speed, a few nops is plenty */
#define WAIT() do { asm volatile("nop; nop; nop;"); } while(0)
#define DLOW() do { PORTC &= ~LCD_DIN; } while(0)
#define DHIGH() do { PORTC |= LCD_DIN; } while(0)
#define CLKLOW() do { PORTC &= ~LCD_CLK; WAIT(); } while(0)
#define CLKHIGH() do { PORTC |= LCD_CLK; WAIT(); } while(0)
#define CSLOW() do { PORTC &= ~LCD_NCS; WAIT(); } while(0)
#define CSHIGH() do { PORTC |= LCD_NCS; WAIT(); } while(0)

static void txByte(uint8_t c)
{
  uint8_t m = 0x80;
  uint8_t i;

  CLKLOW();
  for (i = 0; i < 8; i++) {
    if (c & m) {
      DHIGH();
    } else {
      DLOW();
    }
    WAIT();
    CLKHIGH();
    CLKLOW();
    m = m >> 1;
  }
}

static void txCmd(uint8_t c)
{
  DLOW();
  CLKLOW();
  CSLOW();
  CLKHIGH();    /* L-H latch */
  txByte(c);
  CSHIGH();
}

static void txData(uint8_t d)
{
  DHIGH();
  CLKLOW();
  CSLOW();
  CLKHIGH();    /* L-H latch */
  txByte(d);
  CSHIGH();
}

static void setVOP(uint8_t v)
{
  txCmd(0x20 | (v >> 5));
  txCmd(0x80 | (v & 0x1f));
}

void lcd_init(void)
{
  PORTC |= LCD_NRESET | LCD_NCS;
  DDRC |= LCD_NRESET | LCD_CLK | LCD_NCS | LCD_DIN;

  CLKLOW();
  CSLOW();
  PORTC &= ~LCD_NRESET;
  _delay_ms(20);
  CLKLOW(); CLKHIGH();
  CLKLOW(); CLKHIGH();
  CLKLOW(); CLKHIGH();
  CLKLOW(); CLKHIGH();
  CLKLOW(); CLKHIGH();
  CLKLOW();
  _delay_ms(20);
  PORTC |= LCD_NRESET;
  _delay_ms(20);
  CSHIGH();
  /* PCF8814-ish*/
  txCmd(0xe2); // reset
  _delay_ms(50);

  //  txCmd(0xaa); // vertical addressing (doesn't work! grrr...)
  lcd_on();
  txCmd(0xa4); // dal = 0

  txCmd(0xa6); // normal (non-inverted) display (a7 inverts)
  //  txCmd(0x3a); // osc internal (unnecessary, default?)
  txCmd(0x40); // row = 0
  txCmd(0xb0); // y = 0
  // Bias:
  txCmd(0x37); // no effect, 0-4
  txCmd(0xef); txCmd(0x02); // Frame freq 60Hz
  txCmd(0x10);
  txCmd(0x00);
  //  txCmd(0x3d); txCmd(0x03); // Charge pump x3
  txCmd(0xeb); // temp comp on
  //txCmd(0xea); // temp comp off
  txCmd(0xad);  txCmd(0x81); // N-line inversion
  setVOP(0x88);

  /* Now, invert Y */
  txCmd(0xc8);
}

void lcd_off(void)
{
  txCmd(0xae); // display off
  txCmd(0x28); // CP off
}

void lcd_on(void)
{
  txCmd(0xaf); // display on
  txCmd(0x2f); // CP on (2c didn't work despite 1:0 don'tcare)
}

/* The LCD is in 'horiz addressing' stripes, so need to interleave a
   little to display a 90deg rotated 'linear' image.
 */
void lcd_blit_img_withrotate(char *data)
{
  int8_t row, col;
  unsigned short i;
  row = 0;
  col = 0;
  txCmd(0x10);
  txCmd(0x00); /* x=0 */
  txCmd(0xb0); /* y=0 */

  for (i = 0; i < (96*9); i++) {
    txData(pgm_read_byte(&data[(row * 9) + col]));
    row += 1;
    if (row == 96) {
      row = 0;
      col++;
    }
  }
}

/* Plain for prerotated image */
void lcd_blit_img(char *data)
{
  int8_t row, col;
  unsigned short i;
  row = 0;
  col = 0;
  txCmd(0x10);
  txCmd(0x00); /* x=0 */
  txCmd(0xb0); /* y=0 */

  for (i = 0; i < (96*9); i++) {
    txData(pgm_read_byte(&data[i]));
  }
}

/* For prerotated RLE image */
void lcd_blit_img_rle(char *data, unsigned short len)
{
  int8_t row, col;
  unsigned short i,j;
  row = 0;
  col = 0;
  txCmd(0x10);
  txCmd(0x00); /* x=0 */
  txCmd(0xb0); /* y=0 */

  for (i = 0; i < len; i++) {
    unsigned char b = pgm_read_byte(&data[i]);
    if (b == 0 || b == 0xff) {
      unsigned short l = pgm_read_byte(&data[++i]);
      for (j = 0; j < (l+1); j++)
	txData(b);
    } else {
      txData(b);
    }
  }
}
