/*
  Firmware for Bitmap Projector
  Copyright (C) 2011 Matt Evans

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


  Main code for Bitmap Projector.  Observe IR codes coming in and draw
  images to LCD (possible RLE-compressed).
*/

#include <inttypes.h>

#include "hw.h"          /* CPU clk - define BEFORE delay.h! */
#include "serio.h"
#include "ir.h"
#include "lcd.h"
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>

#include "images/che_rle.h"
#include "images/invaders_rle.h"
#include "images/icons_rle.h"
#include "images/jimi_rle.h"


struct imginfo {
  char *img;
  unsigned short len; /* 0 = non-rle */
} imgs[] = {
  { .img = che_rle, .len = che_rle_len },
  { .img = invaders_rle, .len = invaders_rle_len },
  { .img = icons_rle, .len = icons_rle_len },
  { .img = jimi_rle, .len = jimi_rle_len },
};
#define NUM_IMGS (sizeof(imgs) / sizeof(struct imginfo))  

static void setup(void)
{
  DDRD |= IO_D_LED;
  /* Other cool pins:
     PD7 = fan on/off (high = on)
     PB1 = Lamp on/off (low = on)
  */
  PORTB |= IO_LAMP;
  PORTD &= ~IO_FAN;     /* Both off */
  DDRB |= IO_LAMP;
  DDRD |= IO_FAN;
}

void delay_cs(int n)
{
  int i;
  for (i = 0; i < n; i++)
    _delay_ms(10);
}

int main(void)
{
  uint8_t i = 0;
  uint8_t lcd_pwr = 1; /* Default to on for testing; lcd_init powers up too */

  setup();

  serial_init();
  ir_init();
  lcd_init();
  set_sleep_mode(SLEEP_MODE_IDLE);

  sei(); /* We like IRQs */
  
  PS("---Init done.---\r\n");

  while (1) {
    uint8_t btn = 0;
    uint8_t redisplay = 0;

    sleep_mode();       /* Wait for IRQ, including IR button press. */

    if (ir_new_code) {
      btn = ir_code_to_button(ir_rx);
      ir_new_code = 0;
    }

    if (btn != BTN_UNK) {
      PS("+++ Got IR button ");
      phex8(btn);
      pnewl();
    }
    switch (btn) {
      case BTN_PWR:
        lcd_pwr = !lcd_pwr;
        if (lcd_pwr) {
          FAN_ON();
          delay_cs(20);
          lcd_on();
          LAMP_ON();
          redisplay = 1;
        } else {
          lcd_off();
          LAMP_OFF();
          FAN_OFF();
        }
        break;
      case BTN_VOLMINUS:
        if (--i >= NUM_IMGS)
          i = NUM_IMGS-1;
        redisplay = 1;
        break;
      case BTN_VOLPLUS:
        if (++i >= NUM_IMGS)
          i = 0;
        redisplay = 1;
        break;        
    }

    if (redisplay) {
      if (imgs[i].len)
        lcd_blit_img_rle(imgs[i].img, imgs[i].len);
      else
        lcd_blit_img(imgs[i].img);
      redisplay = 0;
    }
  }
  return 0;
}
